
=pod

=head1 ++ Sub package: crawl

Used for crawling a website by requesting a (start) page, reading the
HTML, extracting the links, and then requesting those links--up to a
specified depth.  The module also allows various configuration tweaks to
do such things as monitor requests for offsite URLs (pages on other
hosts), track various cookies, etc.

=cut

#####################################################
=pod

=head1 - Function: lw::crawl
  
Params: $HOST, $START, $MAX_DEPTH, \%tracking, \%hin
Return: Nothing

The heart of the crawl package.  Will perform an HTTP crawl on the
specified HOST, starting at START URI, proceeding up to MAX_DEPTH.  A
tracking hash reference (required) stores the results of each page (and
ongoing progress).  The optional http_in_options hash reference specifies
a standard HTTP hash for use in the outgoing HTTP requests.  Certain
options are configurable via lw::crawl_set_config().  The tracking hash
will contain all the pages visited; you can get the crawl engine to skip 
pages by placing them in the tracking hash.

=cut
sub crawl {
 my ($HOST, $START, $MAX_DEPTH, $href, $hrin)=@_;
 my (%hout, %jar);
 my (@links, @tlinks, $CWD, @vals, $PORT);

 my $server_tag=undef;
 my (%server_tags, @cookies, @offsites, %offsites);

 @vals=&lw::utils_split_uri($START);
 $START=$vals[0];
 $PORT=$vals[3];

 # some various informationz...
 $crawl_config{'host'}=$HOST;
 $crawl_config{'port'}=$PORT;
 $crawl_config{'start'}=$START;
 $crawl_config{'ref_links'}=\@links;
 $crawl_config{'ref_jar'}=\%jar;
 $crawl_config{'ref_hin'}=\%hin;
 $crawl_config{'ref_hout'}=\%hout;

 push @links, \@{[$START,1,($vals[1] eq 'https')?1:0]};

 while(@links){
  my $C=shift @links;
  my $CURRENT=$C->[0];
  my $DEPTH=$C->[1];
  my $SSL=$C->[2];

  next if($DEPTH > $MAX_DEPTH);
  next if(defined $$href{$CURRENT});

  $$hrin{'whisker'}->{'ssl'} = $SSL;

  $CWD=&lw::utils_get_dir($CURRENT);
  $$hrin{'whisker'}->{'uri'}=$CURRENT;

  if(&lw::http_do_request($hrin,\%hout)){
	# Hmm, error.  What do we do?
	next;
  }

  $$href{$CURRENT}=$hout{'whisker'}->{'http_resp'};

  if($server_tag eq undef){ # different server tag?
   if(defined $hout{'Server'}){ $server_tag=$hout{'Server'};}
  } else {
   if(defined $hout{'Server'} && $hout{'Server'} ne $server_tag){
	$server_tags{$hout{'Server'}}=1;}
  }

  if($hout{'Content-Type'} eq 'text/html' &&
	$hout{'whisker'}->{'http_resp'}==200){

	if(defined $hout{'Set-Cookie'}){
		if($crawl_config{'save_cookies'}>0){
			push @cookies, @{$hout{'whisker'}->{'cookies'}};}
		if($crawl_config{'reuse_cookies'}>0){
			&lw::cookie_read(\%jar,\%hout);
		}
	}

	&lw::html_find_tags(\$hout{'whisker'}->{'data'},
		\&lw::crawl_extract_links_test);

 	foreach $T (@lw::crawl_urls){

	 $T=~tr/\0\r\n//d; # the NULL character is a bug that's somewhere
	 next if (length($T)==0);

	 if($crawl_config{'callback'} != 0){
		next if &{$crawl_config{'callback'}}($T,$CWD); }

	 @vals=lw::utils_split_uri($T);

	 if( (defined $vals[2] && $vals[2] ne $HOST) || 
			$vals[3] != $PORT){
		if($crawl_config{'save_offsites'}>0){
			push @offsites, $T;}
		next; }

	 $vals[0]=~/\.([^.]*)$/;
	 my $EXT = $1 || '';

	 next if($crawl_config{'skip_ext'}=~/\.$EXT /i);

	 if(defined $vals[4] && $crawl_config{'use_params'}>0){
		$vals[0]=$vals[0].'?'.$vals[4];	
	 }

	 $vals[0]=~s#[^/]{0,1}\./##g; # removes '/./'

	 if(substr($vals[0],0,1) ne '/'){
		$vals[0]=$CWD.$vals[0]; }

	 next if(defined $$href{$vals[0]});

	 push @links, \@{[$vals[0],$DEPTH+1, ($vals[1] eq 'https')?1:0]};

	 } # for

	@lw::crawl_urls=(); # reset for next round

  } # if
 } # while

 my $key;
 foreach $key (keys %crawl_config){
 	delete $crawl_config{$key} if $key=~/^ref_/; }

 $crawl_config{'stats_reqs'}=$hout{'whisker'}->{'stats_reqs'};
 $crawl_config{'stats_syns'}=$hout{'whisker'}->{'stats_syns'};

} # end sub crawl

#####################################################
=pod

=head1 - Function: lw::crawl_get_config
  
Params: $config_directive
Return: $config_directive_value

Returns the set value of the submitted config_directive.  See
lw::crawl_set_config() for a list of configuration values.

=cut
sub crawl_get_config {
	return $crawl_config{shift};
}

#####################################################
=pod

=head1 - Function: lw::crawl_set_config
  
Params: $config_directive, $value
Return: Nothing

This function adjusts the configuration of the crawl package. Use values
of 0 and 1 for off and on, respectively.

save_cookies
- crawl will save all cookies encountered, for later review

save_offsite_urls
- crawl will save all offsite URLs (URLs not on this host); crawl
  will not actually crawl those hosts (use separate calls to crawl)

reuse_cookies
- crawl will resubmit any received/prior cookies

skip_ext
- crawl will ignore requests for URLs ending in extensions given; the 
  value requires a specific string format: (dot)extension(space).  For
  example, to ignore GIFs and JPGs, you would run:
 	lw::crawl_set_config('skip_ext',".gif .jpg ");

=cut
sub crawl_set_config {
	$crawl_config{lc(shift)}=$v2;
}

#####################################################

sub crawl_extract_links_test { # tags know to have links
	my ($TAG, $hr, $dr, $start, $len)=(lc(shift),@_);
	my @temp; local $_;

	return if(!scalar %$hr); # fastpath quickie

	while( ($key,$val)= each %$hr){ # normalize element values
#		next if($key eq '');
#		delete $$hr{$key}; # actually don't have to delete
		$$hr{lc($key)} = $val;
	}


	if($TAG eq 'a'){
		if(defined $$hr{'href'}){
			push(@lw::crawl_urls,$$hr{'href'}); }
		return; }

	if($TAG eq 'img'){
		@temp=('src','usemap','lowsrc');
		map { if(defined $$hr{$_}){ 
			push(@lw::crawl_urls,$$hr{$_});}} @temp;
		return; }

	if($TAG eq 'form'){
		if(defined $$hr{'action'}){ 
			push(@lw::crawl_urls,$$hr{'action'});}
		return; }

	if($TAG eq 'applet'){ 
		@temp=('archives','codebase','code');
		map { if(defined $$hr{$_}){ 
			push(@lw::crawl_urls,$$hr{$_});}} @temp;
		return;}

	if($TAG eq 'object'){
		@temp=('classid','codebase','data','archive','usemap');
		map { if(defined $$hr{$_}) {
			push(@lw::crawl_urls,$$hr{$_});}} @temp;
		return;}

	if($TAG eq 'embed'){
		if(defined $$hr{'src'}){ 
			push(@lw::crawl_urls,$$hr{'src'});}
		if(defined $$hr{'pluginspage'}){ 
			push(@lw::crawl_urls,$$hr{'pluginspage'});}
		return;}

	if($TAG eq 'frame'){
		if(defined $$hr{'src'}){ 
			push(@lw::crawl_urls,$$hr{'src'});}
		if(defined $$hr{'longdesc'}){ 
			push(@lw::crawl_urls,$$hr{'longdesc'});}
		return;}

	if($TAG eq 'iframe'){
		if(defined $$hr{'src'}){ 
			push(@lw::crawl_urls,$$hr{'src'});}
		if(defined $$hr{'longdesc'}){ 
			push(@lw::crawl_urls,$$hr{'longdesc'});}
		return;}

	if($TAG eq 'isindex'){
		if(defined $$hr{'action'}){ 
			push(@lw::crawl_urls,$$hr{'action'});}
		return;}

	# leftover tags
	if(defined $$hr{'href'}){ @temp=('area','base','link','xmp');
		map { if($TAG eq $_){
			push(@lw::crawl_urls,$$hr{'href'});}} @temp;}

	if(defined $$hr{'src'}){
		@temp=('bgsound','input','layer','script');
		map { if($TAG eq $_){
			push(@lw::crawl_urls,$$hr{'src'});}} @temp;}

	if(defined $$hr{'background'}){ 
		@temp=('body','ilayer','layer','table','td','th');
		map { if($TAG eq $_){ 
			push(@lw::crawl_urls,$$hr{'background'});
			return; }} @temp;}

	if(defined $$hr{'cite'}){ @temp=('blockquote','del','ins','q');
		map { if($TAG eq $_){ 
			push(@lw::crawl_urls,$$hr{'cite'});
			return; }} @temp;}

	return;}


################################################################

