
=pod

=head1 ++ Sub package: http

The http package is the core package of (lib)whisker.  It is responsible
for making the HTTP requests, and parsing the responses.  It can handle
HTTP 0.9, 1.0, and 1.1 requests, and allows pretty much every aspect of
the request to be configured and controlled.  The HTTP functions use a
HTTP in/out hash, which is a normal perl hash.  For outgoing HTTP requests
('hin' hashes), the keys/values represent outgoing HTTP headers.  For HTTP
responses ('hout' hashes), the keys/values represent incoming HTTP
headers.  For both, however, there is a special key, 'whisker', whose
value is a hash reference.  The whisker control hash contains more
configuration variables, which include host, port, and uri to make the
request for.  To access the whisker control hash, use the following
notation: $hash{'whisker'}->{'key'}='value';

You should view lw::http_init_request() for a list of core whisker control
hash values.

=cut

##################################################################

=pod

=head1 - Function: lw::http_init_request
   
Params: \%input_hash_to_initialize
Return: Nothing (modifies input hash)

Sets default values to the input hash for use.  Sets the host to
'localhost', port 80, request URI '/', using HTTP 1.1 with GET
method.  The timeout is set to 10 seconds, no proxies are defined, and all
URI formatting is set to standard HTTP syntax.  It also sets the
Connection (Keep-Alive) and User-Agent headers.

NOTICE!!  It's important to use http_init_request before calling 
http_do_request, or http_do_request might puke.  Thus, a special magic 
value is placed in the hash to let http_do_request know that the hash has 
been properly initialized.  If you really must 'roll your own' and not use 
http_init_request before you call http_do_request, you will at least need 
to set the INITIAL_MAGIC value (amongst other things).

=cut
sub http_init_request { # doesn't return anything
 my ($hin)=shift;
 $$hin{'whisker'}=(); # clear control hash

# control values
 $$hin{'whisker'}->{'req_spacer'}=' ';
 $$hin{'whisker'}->{'req_spacer2'}=' ';
 $$hin{'whisker'}->{'http_ver'}='1.1'; # notice it is now default!
 $$hin{'whisker'}->{'method'}='GET';
 $$hin{'whisker'}->{'method_postfix'}='';
 $$hin{'whisker'}->{'port'}=80;
 $$hin{'whisker'}->{'uri'}='/';
 $$hin{'whisker'}->{'uri_prefix'}='';
 $$hin{'whisker'}->{'uri_postfix'}='';
 $$hin{'whisker'}->{'uri_param_sep'}='?';
 $$hin{'whisker'}->{'host'}='localhost';
 $$hin{'whisker'}->{'http_req_trailer'}='';
 $$hin{'whisker'}->{'timeout'}=10; # seconds
 $$hin{'whisker'}->{'include_host_in_uri'}=0;
 $$hin{'whisker'}->{'ignore_duplicate_headers'}=1;
 $$hin{'whisker'}->{'normalize_incoming_headers'}=1;
 $$hin{'whisker'}->{'ssl'}=0; # use SSL for this request?
 $$hin{'whisker'}->{'http_eol'}="\r\n";
 $$hin{'whisker'}->{'INITIAL_MAGIC'}=31337;
 
# default header values
 $$hin{'Connection'}='Keep-Alive'; # notice it is now default!
 $$hin{'User-Agent'}='Mozilla/7.3'; # heh
}


##################################################################
=pod

=head1 - Function: lw::http_do_request
   
Params: \%hin, \%hout, \%configs
Return: 1 if error; 0 if no error (also modifies hout hash)

*THE* core function of libwhisker.  lw::http_do_request actually performs
the HTTP request, using the values submitted in %hin, and placing result
values in %hout.  This allows you to resubmit %hin is subsequent requests
(%hout is automatically cleared upon execution).  You can submit 'runtime'
config directives as %configs, which will be spliced into
$hin{'whisker'}->{} before anything else.  That means you can do:

lw::http_do_request(\%hin,\%hout,{'uri'=>'/cgi-bin/'});

This will set $hin{'whisker'}->{'uri'}='/cgi-bin/' before execution, and
provides a simple shortcut (note: it does modify %hin).

Please also take note of the 'hin' and 'hout'.  They are named due to 
what's passed (h)in to the function, and what comes (h)out of the 
function.  It really has nothing to do with the direction of requests 
(which is technically opposite).

=cut
sub http_do_request {
 my ($hin, $hout, $hashref)=@_;
 my ($temp,$vin,$resp,$S,$a,$b,$vout,@c,$c)=(1,'');
 local $_;

 %$hout=(); # clear output hash
 $OUTGOING_QUEUE=$INCOMING_QUEUE='';

 if($LW_HAS_SOCKET==0){
	$$hout{'whisker'}->{'error'}='Socket support not available';
	return 1;}

 if(!defined $$hin{'whisker'} || 
    !defined $$hin{'whisker'}->{'INITIAL_MAGIC'} ||
    $$hin{'whisker'}->{'INITIAL_MAGIC'}!=31337 ){
	$$hout{'whisker'}->{'error'}='Input hash not initialized';
	return 1;
 }

 if(defined $hashref){
	foreach (keys %$hashref){
		$$hin{'whisker'}->{$_}=$$hashref{$_};}}

 if($$hin{'whisker'}->{'ssl'}>0 && $LW_HAS_SSL!=1){
	$$hout{'whisker'}->{'error'}='SSL not available';
	return 1;}

 $TIMEOUT=$$hin{'whisker'}->{'timeout'}||10;

 my $cache_key = defined $$hin{'whisker'}->{'proxy_host'} ?
	"$$hin{'whisker'}->{'proxy_host'}:$$hin{'whisker'}->{'proxy_port'}" :
	"$$hin{'whisker'}->{'host'}:$$hin{'whisker'}->{'port'}" ;

# if(defined $http_host_cache{$cache_key}){
#	$SOCKSTATE=$http_host_cache{$cache_key}


 if($SOCKSTATE>0){
	if($$hin{'whisker'}->{'host'} eq $LAST_HOST){
 		vec($vin,fileno(S),1)=1; # is socket still good?
 		if(!select(undef,$vin,undef,.01)){
			$SOCKSTATE=0;
			sock_close(S,undef);}
	} else { # host changed, so we need new socket
		sock_close(S,undef);
		$SOCKSTATE=0;}   
 }

 if($SOCKSTATE==0){
	if(!socket(S,PF_INET,SOCK_STREAM,getprotobyname('tcp')||0)){
		$$hout{'whisker'}->{'error'}='Socket() problems'; 
		return 1;}

	if($$hin{'whisker'}->{'ssl'}>0){ # ssl setup stuff

	    if($LW_SSL_LIB==1){
		if(! ($SSL_CTX = Net::SSLeay::CTX_new()) ){
			$$hout{'whisker'}->{'error'}="SSL_CTX error: $!";
			return 1;}
		if(! ($SSL_THINGY = Net::SSLeay::new($SSL_CTX)) ){
			$$hout{'whisker'}->{'error'}="SSL_new error: $!";
			return 1;}
	    }
	}

	# use $chost/$cport for actual server we are connecting to
	my ($chost,$cport,$cwhat)=('',80,'');

	if(defined $$hin{'whisker'}->{'proxy_host'}){
	    $chost=$$hin{'whisker'}->{'proxy_host'};
	    $cport=$$hin{'whisker'}->{'proxy_port'};
	    $cwhat='proxy';

	    if($$hin{'whisker'}->{'ssl'}>0 && $LW_SSL_LIB==2){
		$ENV{HTTPS_PROXY} ="$$hin{'whisker'}->{'proxy_host'}:";
		$ENV{HTTPS_PROXY}.=$$hin{'whisker'}->{'proxy_port'}||80; }

	} else {
	    $chost=$$hin{'whisker'}->{'host'};
	    $cport=$$hin{'whisker'}->{'port'};
	    $cwhat='host';
	}

	if($chost ne $LAST_HOST){
		$LAST_HOST=$chost;
		$LAST_INET_ATON=inet_aton($LAST_HOST);}

	if($$hin{'whisker'}->{'ssl'}>0 && $LW_SSL_LIB==2){
		# proxy set in ENV; we always connect to host
		$SSL_THINGY= Net::SSL->new(
			PeerAddr => $$hin{'whisker'}->{'host'},
			PeerPort => $$hin{'whisker'}->{'port'},
			Timeout => $TIMEOUT );
		if($@){ $$hout{'whisker'}->{'error'}="Can't connect via SSL: $@[0]";
			return 1;}
		$SSL_THINGY->autoflush(1);
	} else {
		if(!connect(S,sockaddr_in($cport,$LAST_INET_ATON))){
			$$hout{'whisker'}->{'error'}="Can't connect to host $cwhat";
			return 1;}

		# same as IO::Handle->autoflush(1), without importing 1000+ lines
		$S=select(S); $|++; select($S);
	}

	$SOCKSTATE=1; $STATS_SYNS++;

	if($$hin{'whisker'}->{'ssl'}>0){

	    if($LW_SSL_LIB==1){
		&Net::SSLeay::set_fd($SSL_THINGY, fileno(S));
#		&Net::SSLeay::set_cipher_list($SSL_THINGY, 
#			"DES-CBC3-MD5:RC4-MD5");
		if(! (Net::SSLeay::connect($SSL_THINGY)) ){
			$$hout{'whisker'}->{'error'}="SSL_connect error: $!";
			sock_close(S,$SSL_THINGY); return 1;}
	    }
	} else {
		$SSL_THINGY=undef;
	}
 }

 if(defined $$hin{'whisker'}->{'ids_session_splice'} &&
            $$hin{'whisker'}->{'ids_session_splice'}>0 &&
		$$hin{'whisker'}->{'ssl'}==0){ # no session_spice over ssl
	setsockopt(S,SOL_SOCKET,SO_SNDLOWAT,1);
	@c=split(//, &http_req2line($hin));
	# notice we bypass queueing here, in order to trickle the packets
	foreach $c (@c){ print S $c; select(undef,undef,undef,.05);}
 } else {
	 http_queue(&http_req2line($hin));}

 # send the initial request
 if($res=http_queue_send(S,$SSL_THINGY)){
	$$hout{'whisker'}->{'error'}="Error sending request to server: $res";
	sock_close(S,$SSL_THINGY); $SOCKSTATE=0; return 1;}

 $STATS_REQS++;

 if($$hin{'whisker'}->{'http_ver'} ne '0.9'){

    foreach (keys %$hin){
	next if($_ eq '' || $_ eq 'whisker');
	if(ref($$hin{$_})){ # header with multiple values
		my $key=$_;
		foreach (@{$$hin{$key}}){
		  http_queue("$key: $_$$hin{'whisker'}->{'http_eol'}");}
	} else { # normal header
		http_queue("$_: $$hin{$_}$$hin{'whisker'}->{'http_eol'}");
	}
    }

    if(defined $$hin{'whisker'}->{'raw_header_data'}){
	http_queue($$hin{'whisker'}->{'raw_header_data'});}

    http_queue($$hin{'whisker'}->{'http_eol'});

    if(defined $$hin{'whisker'}->{'data'} && $$hin{'whisker'}->{'meth'}
		ne 'HEAD'){ # PUT, POST, etc
	http_queue($$hin{'whisker'}->{'data'});}

    # all data is wrangled...actually send it now
    if($res=http_queue_send(S,$SSL_THINGY)){
	$$hout{'whisker'}->{'error'}="Error sending data to server: $res";
	sock_close(S,$SSL_THINGY); $SOCKSTATE=0; return 1;}

 } # http 0.9 support

 if(defined $SSL_THINGY){
	if($LW_SSL_LIB==1){ # Net::SSLeay
 		shutdown S, 1; 
	} else { # Net::SSl
		shutdown $SSL_THINGY,1;
	}
 }

 vec($vin,fileno(S),1)=1; # wait only so long to read...
 if(!select($vin,undef,undef,$TIMEOUT)){
	$$hout{'whisker'}->{'error'}="Server read timed out";
	sock_close(S,$SSL_THINGY); $SOCKSTATE=0; return 1;}

if($$hin{'whisker'}->{'http_ver'} ne '0.9'){

 do { # catch '100 Continue' responses
  ($resp=sock_getline(S,$SSL_THINGY))=~tr/\r\n//d;

  if($resp!~/^HTTP\/([0-9.]{3}) (\d+)[ ]{0,1}(.*)/){
	$$hout{'whisker'}->{'error'}="Invalid HTTP response: $resp";
	# let's save the incoming data...we might want it
	$$hout{'whisker'}->{'data'}=$resp;
	while($_=sock_getline(S,$SSL_THINGY)){ 
		$$hout{'whisker'}->{'data'}.=$_;}
	sock_close(S,$SSL_THINGY); $SOCKSTATE=0; # otherwise bad crap lingers
	return 1;}

  ($$hout{'whisker'}->{'http_ver'},$$hout{'whisker'}->{'http_resp'},
	$$hout{'whisker'}->{'http_resp_message'})=($1,$2,$3);

  while($_=sock_getline(S,$SSL_THINGY)){ # check pertinent headers
	tr/\r\n//d; last if ($_ eq '');
	m/^([^:]+): (.*)$/;
	($a,$b)=($1,$2);
	next if ($a eq 'whisker'); # protect ourselves

	if($$hin{'whisker'}->{'normalize_incoming_headers'}>0){
		$a=~s/(-[a-z])/uc($1)/eg;}

	if($a eq 'Set-Cookie'){ # cookies are special
		push(@{$$hout{'whisker'}->{'cookies'}}, $b);
	}

	if(defined $$hout{$a} && 
		$$hin{'whisker'}->{'ignore_duplicate_header'}!=1){
	  if(!ref($$hout{$a})){
	    my $temp=$$hout{$a};
	    delete $$hout{$a};
	    push(@{$$hout{$a}},$temp);
	  }
	  push(@{$$hout{$a}},$b);
	} else {
	  $$hout{$a}=$b;
  }	}

 } while($$hout{'whisker'}->{'http_resp'}==100);

} else { # http ver 0.9, we need to fake it
 $$hout{'whisker'}->{'http_ver'}='0.9';
 $$hout{'whisker'}->{'http_resp'}='200';
 $$hout{'whisker'}->{'http_resp_message'}='';
}

 if($$hin{'whisker'}->{'method'} ne 'HEAD' && 
	$$hout{'whisker'}->{'http_resp'}!=206 &&
	$$hout{'whisker'}->{'http_resp'}!=102){
  if (defined $$hout{'Transfer-Encoding'} &&
	      $$hout{'Transfer-Encoding'} eq 'chunked') { 
	$a=sock_getline(S,$SSL_THINGY);
	$$hout{'Content-Length'}=hex($a); 
	while($$hout{'Content-Length'}!=0) { # chunked sucks
		if(!defined ($temp=sock_get(S,$SSL_THINGY,
				$$hout{'Content-Length'}))){
			$$hout{'whisker'}->{'error'}='Error reading data';
			sock_close(S,$SSL_THINGY); $SOCKSTATE=0; return 1;}
		$$hout{'whisker'}->{'data'}=$$hout{'whisker'}->{'data'} . $temp;
		$temp=sock_getline(S, $SSL_THINGY);
		($temp=sock_getline(S, $SSL_THINGY)) if($temp=~/^[\r\n]*$/);
		$$hout{'Content-Length'}=hex($temp);}
	while($_=sock_getline(S,$SSL_THINGY)){ tr/\r\n//d; last if($_ eq ''); }
  } else {
 	if (defined $$hout{'Content-Length'}) {
		if(!defined ($temp=sock_get(S,$SSL_THINGY,
				$$hout{'Content-Length'}))){
			$$hout{'whisker'}->{'error'}='Error reading data';
			sock_close(S,$SSL_THINGY); $SOCKSTATE=0; return 1;}
	} else { $temp=''; while($_=sock_getline(S,$SSL_THINGY)){$temp.=$_;}}
	$$hout{'whisker'}->{'data'}=$temp; 
  }
 } # /method ne HEAD && http_resp ne 206 or 102/

 if(!defined $$hout{'Connection'} || $$hout{'Connection'} ne 'Keep-Alive') {
	$$hout{'Connection'}='Close';
	$SOCKSTATE=0;
	sock_close(S,$SSL_THINGY); } 

 $$hout{'whisker'}->{'sockstate'}=$SOCKSTATE;
 $$hout{'whisker'}->{'stats_reqs'}=$STATS_REQS;
 $$hout{'whisker'}->{'stats_syns'}=$STATS_SYNS;
 $$hout{'whisker'}->{'error'}=''; # no errors
 return 0;
}


##################################################################
=pod

=head1 - Function: lw::http_req2line
  
Params: \%hin, $switch
Return: $request

req2line is used internally by lw::http_do_request, as well as provides a
convienient way to turn a %hin configuration into an actual HTTP request
line.  If $switch is set to 1, then the returned $request will be the URI
only ('/requested/page.html'), versus the entire HTTP request ('GET
/requested/page.html HTTP/1.0\n\n').  Also, if the 'full_request_override'
whisker config variable is set in %hin, then it will be returned instead
of the constructed URI.

=cut
sub http_req2line {
 my ($S,$hin,$UO)=('',@_);
 $UO||=0; # shut up -w warning

 # notice: full_request_override can play havoc with proxy settings
 if(defined $$hin{'whisker'}->{'full_request_override'}){
	return $$hin{'whisker'}->{'full_request_override'};

 } else { # notice the components of a request--this is for flexibility

	if($UO!=1){$S.= 	$$hin{'whisker'}->{'method'}.
				$$hin{'whisker'}->{'method_postfix'}.
				$$hin{'whisker'}->{'req_spacer'};
	
		if($$hin{'whisker'}->{'include_host_in_uri'}>0){
			$S.=	'http://';

			if(defined $$hin{'whisker'}->{'uri_user'}){
			$S.=	$$hin{'whisker'}->{'uri_user'};
			if(defined $$hin{'whisker'}->{'uri_password'}){
				$S.=	':'.$$hin{'whisker'}->{'uri_user'};
			}
			$S.=	'@';
			}

			$S.=	$$hin{'whisker'}->{'host'}.
				':'.$$hin{'whisker'}->{'port'};}}

	$S.=	$$hin{'whisker'}->{'uri_prefix'}.
		$$hin{'whisker'}->{'uri'}.
		$$hin{'whisker'}->{'uri_postfix'};

	if(defined $$hin{'whisker'}->{'uri_param'}){
		$S.= 	$$hin{'whisker'}->{'uri_param_sep'}.
			$$hin{'whisker'}->{'uri_param'};}

	if($UO!=1){  if($$hin{'whisker'}->{'http_ver'} ne '0.9'){
			$S.= 	$$hin{'whisker'}->{'req_spacer2'}.'HTTP/'.
				$$hin{'whisker'}->{'http_ver'}.
				$$hin{'whisker'}->{'http_req_trailer'};}
			$S.=	$$hin{'whisker'}->{'http_eol'};}}
 return $S;}



##################################################################
=pod

=head1 - Function lw::sock_close
   
Params: $socket_file_descriptor, $SSL_THINGY
Return: nothing

This function will close the indicated socket and SSL connection (if 
necessary).  They are wrapped in eval()s to make sure if the functions 
puke, it doesn't kill the entire program.

=cut
sub sock_close {
	my ($fd,$ssl)=@_;

	eval "close($fd)";
	if(defined $ssl){
	    if($LW_SSL_LIB==1){ # Net::SSLeay
		eval "&Net::SSLeay::free($ssl)";
	    } else { # Net::SSL
		eval "close($SSL_THINGY)"; # is that right for Net::SSL?
	    }
	}

	$SSL_THINGY=undef;
}


##################################################################
=pod

=head1 - Function: lw::sock_getline
   
Params: $socket_file_descriptor, $SSL_THINGY
Return: $string, undef on error (timeout)

This function is used internally to read a line of input (up to a '\n')
from the given socket file descriptor (regular or SSL).

This function is not intended for external use.

=cut
sub sock_getline { # read from socket w/ timeouts
        my ($fd,$ssl) = @_;
        my ($str,$t)=('','');

        $t = index($INCOMING_QUEUE,"\n",0);

        while($t < 0){
                return undef if &http_queue_read($fd,$ssl);
                $t=index($INCOMING_QUEUE,"\n",0);
        }

        return substr($INCOMING_QUEUE,0,$t+1,'');
}

##################################################################
=pod

=head1 - Function: lw::sock_get
   
Params: $socket_file_descriptor, $SSL_THINGY, $amount
Return: $string, undef on error

This function is used internally to read input from the given socket 
file descriptor (regular or SSL).

This function is not intended for external use.

=cut
sub sock_get { # read from socket w/ timeouts
        my ($fd,$ssl,$amount) = @_;
        my ($str,$t)=('','');

	while($amount > length($INCOMING_QUEUE)){
                return undef if &http_queue_read($fd,$ssl);
	}

	return substr($INCOMING_QUEUE,0,$amount,'');
}


##################################################################
=pod

=head1 - Function: lw::http_queue_read
   
Params: $fd, $ssl
Return: $character, undef on error (timeout)

http_queue_read() will return one character from the incoming queue
of data.  If the queue is empty, it attempts to read more data from
the server.  If there's no more data, it will return undef.

This function is really for internal use only.

=cut
sub http_queue_read {
	my ($fd,$ssl)=@_;
	my ($vin, $t)=('');

	if(defined $ssl){
	    if($LW_SSL_LIB==1){ # Net::SSLeay
        	if(!($INCOMING_QUEUE.=Net::SSLeay::ssl_read_all($ssl))){
			return 1;}
	    } else { # Net::SSL
		if(!$SSL_THINGY->read($t,1024)){ return 1;
		} else { $INCOMING_QUEUE.=$t;}
	    }
	} else {
		vec($vin,fileno($fd),1)=1; # wait only so long to read...
		if(!select($vin,undef,undef,$TIMEOUT)){
			return 1;}
               	if(!sysread($fd,$t,1024)){	return 1;
		} else {			$INCOMING_QUEUE.=$t;}
	}

	return 0;
}

##################################################################
=pod

=head1 - Function: lw::http_queue_send
   
Params: $sock, $ssl
Return: $status_result (undef=ok, else error message)

This functions sends the current queue (made with lw::http_queue) to the 
server via the specified SSL or socket connection.

=cut
sub http_queue_send { # write to socket
	my ($fd,$ssl)=@_;
	my ($v,$wrote,$error)=('');

	if(defined $ssl){
	    if($LW_SSL_LIB==1){ # Net::SSLeay
		($wrote,$err)=Net::SSLeay::ssl_write_all($ssl,$OUTGOING_QUEUE);
		return "SSL_write error: $err" unless $wrote;
	    } else { # Net::SSL
		$SSL_THINGY->print($OUTGOING_QUEUE);
	    }
	} else {
        	vec($v,fileno($fd),1)=1;
 		if(!select(undef,$v,undef,.01)){ 
			return 'Socket write test failed'; }
		print $fd $OUTGOING_QUEUE;
	}
	$OUTGOING_QUEUE=''; return undef;
}


##################################################################
=pod

=head1 - Function: lw::http_queue
   
Params: $data
Return: nothing

This function will buffer the output to be sent to the server.  Output is 
buffered for various reasons (particularlly because of SSL, but also 
allowing the chance to 'go back' and modify the final output before it's 
actually sent (after header constructions, etc).

=cut
sub http_queue {
	$OUTGOING_QUEUE.= shift;
}


##################################################################
=pod

=head1 - Function: lw::http_fixup_request
   
Params: $hash_ref
Return: Nothing

This function takes a %hout hash reference and makes sure the proper 
headers exist (for example, it will add the Host: header, calculate the 
Content-Length: header for POST requests, etc).  For standard requests 
(i.e. you want the request to be HTTP-RFC compliant), you should call this 
function right before you call lw::http_do_request.

=cut
sub http_fixup_request {
 my $hin=shift;

 if($$hin{'whisker'}->{'http_ver'} eq '1.1' && !defined $$hin{'Host'}){
	$$hin{'Host'}=$$hin{'whisker'}->{'host'};}

 if(defined $$hin{'whisker'}->{'data'} && $$hin{'whisker'}->{'meth'} ne 'HEAD'
		&& !defined $$hin{'Content-Length'}){
	$$hin{'Content-Length'}=length($$hin{'whisker'}->{'data'});}

 if(defined $$hin{'whisker'}->{'proxy_host'}){
	$$hin{'whisker'}->{'include_host_in_uri'}=1;}

}

##################################################################

